"""
Reference
https://dfo-ocean-navigator.github.io/Ocean-Navigator-Manual/
"""


import requests
import os
import shutil
from urllib.request import urlopen
from urllib.parse import urlencode
from contextlib import closing
try:
   from PIL import Image
except:
   print("If you are on a Windows machine, please install PIL (imaging library) using 'python -m pip install Pillow' in the Anaconda Prompt")
   exit()
import json

def requestFile():
   # json object for the query
   query = {"area":[{"innerrings":[],
"name":"",
"polygons":[[[45.06576154770312,-61.17187499999999],[43.60426186809619,-64.951171875],[41.153842357114456,-64.79736328124999],[42.33418438593941,-60.84228515624999],[43.779026621608295,-61.61132812499999],[44.535674532413196,-60.46875],[45.06576154770312,-61.17187499999999]]]}],
"bathymetry":1,
"colormap":"default",
"contour":{"colormap":"default",
"hatch":0,
"legend":1,
"levels":"auto",
"variable":"none"},
"dataset":"giops_day",
"depth":"11",
"interp":"gaussian",
"neighbours":10,
"projection":"EPSG:3857",
"quantum":"day",
"quiver":{"colormap":"default",
"magnitude":"length",
"variable":"none"},
"radius":25,
"scale":"30,40,auto",
"showarea":1,
"time":2211451200,
"type":"map",
"variable":"vosaline"}
   # Assemble full request
   base_url = "http://navigator.oceansdata.ca/api/v1.0/plot/?"
   url = base_url + urlencode({"query": json.dumps(query)}) + '&save&format=csv&size=10x7&dpi=144'
   print(url)
   # Save file and finish
   data_file = requests.get(url, stream=True)
   dump = data_file.raw
   # change this if you want a different save location
   location = os.getcwd()
   with open("script_output.csv", "wb") as location:
      print('Saving File')
      shutil.copyfileobj(dump, location)
      print('Done')

if __name__ == '__main__':
   requestFile()



import logging
from PIL import Image
from functools import reduce
from xml.etree import ElementTree as ET
import json

import matplotlib
matplotlib.use('qt5agg')
import mpl_scatter_density
import matplotlib.pyplot as plt
import netCDF4
import numpy as np

from kadlu.geospatial.data_sources.data_util        import          \
        database_cfg,                                               \
        storage_cfg,                                                \
        insert_hash,                                                \
        serialized,                                                 \
        index


def load_raster(filepath, plot=False, cmap=None, **kwargs):
    """ load data from raster file 

        args:
            filepath: string
                complete filepath descriptor of netcdf file to be read
            plot: boolean
                if True, a plot will be displayed using the qt5agg backend
            cmap: matplotlib colormap
                the full list of available colormaps can be viewed with:
                print(matplotlib.pyplot.colormaps())
                if None is supplied, pyplot will default to 
                matplotlib.pyplot.cm.cividis

        returns:
            values: numpy 2D array
            lats:   numpy 1D array
            lons:   numpy 1D array
    """

    # load raster
    Image.MAX_IMAGE_PIXELS = 500000000
    im = Image.open(filepath)

    # GDAL raster format
    # http://duff.ess.washington.edu/data/raster/drg/docs/geotiff.txt
    if 33922 in im.tag.tagdata.keys():
        i,j,k,x,y,z = im.tag_v2[33922]  # ModelTiepointTag
        dx, dy, dz  = im.tag_v2[33550]  # ModelPixelScaleTag
        meta        = im.tag_v2[42112]  # GdalMetadata
        xml         = ET.fromstring(meta)
        params      = {tag.attrib['name'] : tag.text for tag in xml}
        lat = np.arange(y, y + (dy * im.size[1]), dy)[ :: -1] -90
        rng_lat = (abs(index(kwargs['north'], lat[::-1])-len(lat)), 
                   abs(index(kwargs['south'], lat[::-1])-len(lat)))
        lon = np.arange(x, x + (dx * im.size[0]), dx)
        rng_lon = index(kwargs['east'], lon), index(kwargs['west'], lon)
        logging.info(f'{xml.tag}\nraster coordinate system: {im.tag_v2[34737]}'
                     f'\n{json.dumps(params, indent=2, sort_keys=True)}')

    # NASA / jet propulsion labs raster format (page 27)
    # https://landsat.usgs.gov/sites/default/files/documents/geotiff_spec.pdf
    elif 34264 in im.tag.tagdata.keys():
        dx,_,_,x,_,dy,_,y,_,_,dz,z,_,_,_,_ = im.tag_v2[34264]  # ModelTransformationTag
        lat = np.arange(y, y + (dy * im.size[1]), dy)
        rng_lat = index(kwargs['south'], -lat),  index(kwargs['north'], -lat)
        lon = np.arange(x, x + (dx * im.size[0]), dx)
        rng_lon = index(kwargs['west'], lon), index(kwargs['east'], lon)

    else: assert False, 'unknown metadata tag encoding'
    assert not (z or dz), '3D rasters not supported yet'

    # construct grid and decode pixel values
    if reduce(np.multiply, (rng_lon[1] - rng_lon[0], rng_lat[1] - rng_lat[0])) > 10000000: 
        logging.info('this could take a few moments...')
    grid = np.ndarray((len(np.arange(rng_lon[0], rng_lon[1])), len(np.arange(rng_lat[0], rng_lat[1]))))
    for yi in np.arange(rng_lon[0], rng_lon[1]) -rng_lon[0]: 
        grid[yi] = np.array(list(map(im.getpixel, zip(
            map(int, [yi for xi in np.arange(rng_lon[0], rng_lon[1]) -rng_lon[0]]), 
            map(int,               np.arange(rng_lat[0], rng_lat[1]) -rng_lat[0] )
        ))))
    mask = grid == float(im.tag_v2[42113])
    val = np.ma.MaskedArray(grid, mask=mask)

    # plot the data
    if plot:
        x1, y1 = np.meshgrid(lon[rng_lon[0]:rng_lon[1]], lat[rng_lat[0]:rng_lat[1]], indexing='ij')
        fig = plt.figure()
        if (rng_lon[1]-rng_lon[0]) * (rng_lat[1]-rng_lat[0]) >= 100000:
            ax = fig.add_subplot(1,1,1, projection='scatter_density')
            plt.axis('scaled')
            raster = ax.scatter_density(x1, y1, c=val, cmap=cmap)
            plt.tight_layout()
        else:
            ax = fig.add_subplot(1,1,1)
            ax.scatter(x1, y1, c=val, cmap=cmap)
        plt.show()

    #return val, y1, x1
    return val, lat[rng_lat[0]:rng_lat[1]], lon[rng_lon[0]:rng_lon[1]]


def load_netcdf(filename, var=None, plot=False, cmap=None, **kwargs):
    """ read environmental data from netcdf and output to gridded numpy array

        args:
            filename: string
                complete filepath descriptor of netcdf file to be read
            var: string (optional)
                the netcdf attribute to be read as the values.
                by default, a guess will be made based on the file metadata
            plot: boolean
                if True, a plot will be displayed using the qt5agg backend
            cmap: matplotlib colormap
                the full list of available colormaps can be viewed with:
                print(matplotlib.pyplot.colormaps())
                if None is supplied, pyplot will default to 
                matplotlib.pyplot.cm.cividis

        returns:
            values: numpy 2D array
            lats:   numpy 1D array
            lons:   numpy 1D array
    """

    ncfile = netCDF4.Dataset(filename)

    if var is None:
        assert 'lat' in ncfile.variables.keys()
        assert 'lon' in ncfile.variables.keys()
        assert len(ncfile.variables.keys()) == 3
        var = [_ for _ in ncfile.variables.keys() if _ != "lat" and _ != "lon"][0]

    assert var in ncfile.variables, f'variable {var} not in file. file contains {ncfile.variables.keys()}'

    logging.info(f'loading {var} from {ncfile.getncattr("title")}')

    rng_lat = index(kwargs['west'],  ncfile['lat'][:].data), index(kwargs['east'],  ncfile['lat'][:].data)
    rng_lon = index(kwargs['south'], ncfile['lon'][:].data), index(kwargs['north'], ncfile['lon'][:].data)

    val = ncfile[ var ][:].data[rng_lat[0]:rng_lat[1], rng_lon[0]:rng_lon[1]]
    lat = ncfile['lat'][:].data[rng_lat[0]:rng_lat[1]]
    lon = ncfile['lon'][:].data[rng_lon[0]:rng_lon[1]]

    # plot the data
    if plot:
        x1, y1 = np.meshgrid(lon[rng_lon[0]:rng_lon[1]], lat[rng_lat[0]:rng_lat[1]], indexing='ij')
        fig = plt.figure()
        if (rng_lon[1]-rng_lon[0]) * (rng_lat[1]-rng_lat[0]) >= 100000:
            ax = fig.add_subplot(1,1,1, projection='scatter_density')
            plt.axis('scaled')
            raster = ax.scatter_density(x1, y1, c=val, cmap=cmap)
            plt.tight_layout()
        else:
            ax = fig.add_subplot(1,1,1)
            ax.scatter(x1, y1, c=val, cmap=cmap)
        plt.show()

    return val, lat, lon




import os

# data utils
from .geospatial.data_sources.data_util import \
        era5_cfg,       \
        database_cfg,   \
        storage_cfg,    \
        epoch_2_dt,     \
        dt_2_epoch,     \
        index,          \
        reshape_2D,     \
        reshape_3D

# loading with automatic fetching
from .geospatial.data_sources.source_map import source_map 
from .geospatial.data_sources.chs import Chs as chs
from .geospatial.data_sources.era5 import Era5 as era5
from .geospatial.data_sources.gebco import Gebco as gebco
from .geospatial.data_sources.hycom import Hycom as hycom
from .geospatial.data_sources.wwiii import Wwiii as wwiii

# load data from local files
from .geospatial.data_sources.load_from_file import load_netcdf
from .geospatial.data_sources.load_from_file import load_raster

# user-facing data loading API
from .geospatial.data_sources.source_map import load_map


def load(source, var, **kwargs):
    """ automated fetching and loading from web sources 

        args
            source, var: strings
                to view the complete list of sources and variables:
                print(kadlu.source_map)

            kwargs: dictionary
                dict containing boundary coordinates. example:
                kwargs=dict(
                    south=44.25, west=-64.5,
                    north=44.70, east=-63.33,
                    top=0, bottom=5000,
                    start=datetime(2015, 3, 1), end=datetime(2015, 3, 1, 12)
                )

        returns an ND numpy array. arrays ordered by:
            val, lat, lon, [time, depth]
            times are in epoch format
    """

    source, var = source.lower(), var.lower()
    if var == 'bathymetry' or var == 'depth' or var == 'elevation': var = 'bathy'

    loadkey = f'{var}_{source}'
    assert loadkey in load_map.keys(), 'invalid source or variable. valid options include: \n\n'\
            f'{list(f.rsplit("_", 1)[::-1] for f in load_map.keys())}\n\n'\
            f'for more info, print(kadlu.source_map)'

    return load_map[loadkey](**kwargs)


def load_file(filepath, **kwargs):
    """ loading from local files 

        args
            filepath: string
                full path directory and filename of data to load

            kwargs: dictionary
                dict containing coordinate bounds, e.g. north, south, 
                east, west, top, bottom, start, end
                note that times are in datetime format

        returns an ND numpy array. arrays ordered by:
            val, lat, lon, [time, depth]
            times are in epoch format
    """
    assert os.path.isfile(filepath), f'could not find {filepath}'
    ext = lambda filepath, extensions: isinstance(extensions, tuple) and any(ext == filepath.lower()[-len(ext):] for ext in extensions)

    if ext(filepath, ('.nc',)):
        return load_netcdf(filepath, **kwargs)

    elif ext(filepath, ('.tif', '.tiff', '.gtiff',)):       
        return load_raster(filepath, **kwargs)

    else:
        assert False, 'unknown file format - currently only .nc and .tiff formats are accepted'




import os
import logging
import zipfile
import requests

from kadlu.geospatial.data_sources.load_from_file import load_netcdf
from kadlu.geospatial.data_sources.data_util        import          \
        storage_cfg,                                                \
        insert_hash,                                                \
        serialized

class Gebco():

    def fetch_bathymetry(self, **kwargs):
        """ fetch gebco netcdf bathymetry, and return the filepath of extracted data """

        if not serialized(seed='gebco_bathy.nc'):
            logging.info('downloading and decompressing gebco bathymetry from netcdf (~8GB)... ')

            # download zipped netcdf to storage directory
            url = 'https://www.bodc.ac.uk/data/open_download/gebco/gebco_2020/zip/'
            with requests.get(url, stream=True) as payload_netcdf:
                assert payload_netcdf.status_code == 200, 'error fetching file'
                with open(storage_cfg()+'gebco_netcdf.zip', 'wb') as f:
                    for chunk in payload_netcdf.iter_content(chunk_size=8192): 
                        f.write(chunk)

            # unzip it
            with zipfile.ZipFile(storage_cfg()+'gebco_netcdf.zip', 'r') as zipf:
                zipf.extractall(storage_cfg())
            unzipped = zipfile.ZipFile(storage_cfg()+"gebco_netcdf.zip", "r").namelist()
            ncpath = [_ for _ in unzipped if _[-3:] == '.nc'][0]
            os.rename(storage_cfg() + ncpath, storage_cfg() + 'gebco_bathy.nc')
            renamed = [fname if fname[-3:] != ".nc" else "gebco_bathy.nc" for fname in unzipped]

            # store some metadata
            insert_hash(seed='gebco_bathy.nc')

            logging.info(f'extracted {renamed} to {storage_cfg()}')

        return storage_cfg() + 'gebco_bathy.nc'


    def load_bathymetry(self, **kwargs):
        val, lat, lon = load_netcdf(filename=self.fetch_bathymetry(), **kwargs)
        return val * -1, lat, lon


